DownRToS2dv <- function(...) {
  library(abind)

  args <- list(...)
  if (!all(sapply(args, 
                  function(x) {
                    all(c('Data', 'Variable', 'Dates', 'xyCoords') %in% names(x)) &&
                    all(c('dataset', 'source', 'URL') %in% names(attributes(x)))
                  }))) {
    stop("All provided objects in parameter '...' must be full dowsncaleR objects.")
  }

  is_exp <- sapply(args, function(x) 'InitializationDates' %in% names(x))
  exp_indices <- which(is_exp)
  obs_indices <- which(!is_exp)

  provided_data <- list(exp = list(), obs = list())

  # Populate provided_data with info on provided experimental data
  found_sdates <- NULL
  for (exp in exp_indices) {
    dataset <- attr(args[[exp]], 'dataset')
    if (!(dataset %in% names(provided_data[['exp']]))) {
      provided_data[['exp']][[dataset]] <- list()
    }
    if (is.null(found_sdates)) {
      found_sdates <- args[[exp]][['InitializationDates']][[1]]
    } else if (!identical(args[[exp]][['InitializationDates']][[1]], found_sdates)) {
      stop("Provided downscaleR objects with different sets of start dates. ",
           "Objects with data for only one set of start dates must be provided.")
    }
    exp_members <- args[[exp]][['Members']]
    found_members <- lapply(provided_data[['exp']][[dataset]], '[[', 'members')
    matches <- sapply(found_members, identical, exp_members)
    if (any(matches)) {
      exp_ftimes <- args[[exp]]$Dates$start
      found_ftimes <- lapply(provided_data[['exp']][[dataset]][[which(matches)[1]]][['chunks']], '[[', 'ftimes')
      ftime_matches <- sapply(found_ftimes, identical, exp_ftimes)
      if (any(ftime_matches)) {
        stop("Provided twice the data for the same experiment, member and range of forecast times.")
      } else {
        provided_data[['exp']][[dataset]][[which(matches)[1]]][['chunks']] <- c(provided_data[['exp']][[dataset]][[which(matches)[1]]][['chunks']], 
                                                                                list(
                                                                                  list(ftimes = args[[exp]]$Dates$start,
                                                                                       exp_index = exp
                                                                                      )
                                                                                    )
                                                                               )
      }
    } else {
      provided_data[['exp']][[dataset]] <- c(provided_data[['exp']][[dataset]], 
                                             list(
                                               list(members = exp_members, 
                                                    chunks = list(
                                                               list(ftimes = args[[exp]]$Dates$start,
                                                                    exp_index = exp
                                                                   )
                                                                 )
                                                   )
                                                 )
                                            )
    }
  }

  # Populate provided_data with info on provided observational data.
  for (obs in obs_indices) {
    dataset <- attr(args[[obs]], 'dataset')
    if (!(dataset %in% names(provided_data[['obs']]))) {
      provided_data[['obs']][[dataset]] <- list()
    }
    obs_members <- 'Member_1'
    found_members <- lapply(provided_data[['obs']][[dataset]], '[[', 'members')
    matches <- sapply(found_members, identical, obs_members)
    if (any(matches)) {
      obs_ftimes <- args[[obs]]$Dates$start
      found_ftimes <- lapply(provided_data[['obs']][[dataset]][[which(matches)[1]]][['chunks']], '[[', 'ftimes')
      ftime_matches <- sapply(found_ftimes, identical, obs_ftimes)
      if (any(ftime_matches)) {
        stop("Provided twice the data for the same observation and range of forecast times.")
      } else {
        provided_data[['obs']][[dataset]][[which(matches)[1]]][['chunks']] <- c(provided_data[['obs']][[dataset]][[which(matches)[1]]][['chunks']], 
                                                                                list(
                                                                                  list(ftimes = args[[obs]]$Dates$start,
                                                                                       obs_index = obs
                                                                                      )
                                                                                    )
                                                                               )
      }
    } else {
      provided_data[['obs']][[dataset]] <- c(provided_data[['obs']][[dataset]], 
                                             list(
                                               list(members = obs_members, 
                                                    chunks = list(
                                                               list(ftimes = args[[obs]]$Dates$start,
                                                                    obs_index = obs
                                                                   )
                                                                 )
                                                   )
                                                 )
                                            )
    }
  }

  # THIS CODE IS ASSUMING NO OVERLAPPING SDATES COME FROM DOWSNCALER.
  # Casting a joint downscaleR object with all experimental data for each exp.
  downr_exps <- list()
  for (exp_name in names(provided_data$exp)) {
    for (member in 1:length(provided_data$exp[[exp_name]])) {
      new_member <- NULL
      for (chunk in 1:length(provided_data$exp[[exp_name]][[member]][['chunks']])) {
        if (!is.null(new_member)) {
          new_member <- bindGrid.time(new_member,
                                      args[[provided_data$exp[[exp_name]][[member]][['chunks']][[chunk]][['exp_index']]]])
        } else {
          new_member <- args[[provided_data$exp[[exp_name]][[member]][['chunks']][[chunk]][['exp_index']]]]
        }
      }
      if (exp_name %in% names(downr_exps)) {
        ## TODO: If dates do not match with existing members or with other exps, extend downscaleR objects with NAs.
        downr_exps[[exp_name]] <- bindGrid.member(downr_exps[[exp_name]], new_member)
      } else {
        downr_exps[[exp_name]] <- new_member
      }
    }
  }
  # Casting a joint downscaleR object with all experimental data for each exp.
  downr_exps <- list()
  for (exp_name in names(provided_data$exp)) {
    for (member in 1:length(provided_data$exp[[exp_name]])) {
      new_member <- NULL
      for (chunk in 1:length(provided_data$exp[[exp_name]][[member]][['chunks']])) {
        if (!is.null(new_member)) {
          new_member <- bindGrid.time(new_member,
                                      args[[provided_data$exp[[exp_name]][[member]][['chunks']][[chunk]][['exp_index']]]])
        } else {
          new_member <- args[[provided_data$exp[[exp_name]][[member]][['chunks']][[chunk]][['exp_index']]]]
        }
      }
      if (exp_name %in% names(downr_exps)) {
        ## TODO: If dates do not match with existing members or with other exps, extend downscaleR objects with NAs.
        downr_exps[[exp_name]] <- bindGrid.member(downr_exps[[exp_name]], new_member)
      } else {
        downr_exps[[exp_name]] <- new_member
      }
    }
  }

  names(dim(data_exp$Data)) <- attr(data_exp$Data, 'dimensions')
  names(dim(data_obs$Data)) <- attr(data_obs$Data, 'dimensions')
  ## TODO: compute start_indices by doing a substraction actual_date - init_dates and picking the smallest positive distance.
  start_indices <- which(data_exp$Dates$start %in% gsub("11-01", "12-01", data_exp$InitializationDates))
  total_dates <- length(data_exp$Dates$start)
  ftime_amounts <- c(start_indices, total_dates + 1)[2:(length(start_indices) + 1)] - start_indices
  largest_ftime <- max(ftime_amounts)
  # Reordering data
  new_data_exp <- data_exp$Data
  new_data_obs <- data_obs$Data
  new_dates <- data_exp$Dates
  chunk_dims_exp <- dim(data_exp$Data)
  chunk_dims_obs <- dim(data_obs$Data)
  for (i in 1:length(ftime_amounts)) {
    ftime_amount <- ftime_amounts[i]
    if (ftime_amount < largest_ftime) {
      # Fill in exp array and dates
      step_new_data_exp <- NULL
      step_new_dates_s <- NULL
      step_new_dates_e <- NULL
      #if (largest_ftime * (i - 1) + ftime_amount > 0) {
      step_new_data_exp <- Subset(new_data_exp, 'time', 1:(largest_ftime * (i - 1) + ftime_amount))
      step_new_dates_s <- new_dates$start[1:(largest_ftime * (i - 1) + ftime_amount)]
      step_new_dates_e <- new_dates$end[1:(largest_ftime * (i - 1) + ftime_amount)]
      #}
      chunk_dims_exp['time'] <- largest_ftime - ftime_amount
      step_new_data_exp <- abind(step_new_data_exp, array(dim = chunk_dims_exp), along = which(names(chunk_dims_exp) == 'time'))
      names(dim(step_new_data_exp)) <- names(chunk_dims_exp)
      step_new_dates_s <- c(step_new_dates_s, rep(NA, chunk_dims_exp['time']))
      step_new_dates_e <- c(step_new_dates_e, rep(NA, chunk_dims_exp['time']))
      if (dim(new_data_exp)[which(names(chunk_dims_exp) == 'time')] > (largest_ftime * (i - 1) + ftime_amount)) {
        step_new_data_exp <- abind(step_new_data_exp, 
                               Subset(new_data_exp, 'time', (largest_ftime * (i - 1) + ftime_amount + 1):
                                                        (dim(new_data_exp)[which(names(chunk_dims_exp) == 'time')])),
                               along = which(names(chunk_dims_exp) == 'time'))
        names(dim(step_new_data_exp)) <- names(chunk_dims_exp)
        step_new_dates_s <- c(step_new_dates_s, 
                              new_dates$start[(largest_ftime * (i - 1) + ftime_amount + 1):
                                              (dim(new_data_exp)[which(names(chunk_dims_exp) == 'time')])]
                             )
        step_new_dates_e <- c(step_new_dates_e, 
                              new_dates$end[(largest_ftime * (i - 1) + ftime_amount + 1):
                                            (dim(new_data_exp)[which(names(chunk_dims_exp) == 'time')])]
                             )
      }
      new_dates <- list(start = step_new_dates_s,
                        end = step_new_dates_e)
      new_data_exp <- step_new_data_exp
      rm(step_new_data_exp)
      rm(step_new_dates_s)
      rm(step_new_dates_e)
      gc()
      # Fill in obs array
      step_new_data_exp <- NULL
      if (largest_ftime * (i - 1) + ftime_amount > 0) {
        step_new_data_exp <- Subset(new_data_obs, 'time', 1:(largest_ftime * (i - 1) + ftime_amount))
      }
      chunk_dims_obs['time'] <- largest_ftime - ftime_amount
      step_new_data_exp <- abind(step_new_data_exp, array(dim = chunk_dims_obs), along = which(names(chunk_dims_obs) == 'time'))
      names(dim(step_new_data_exp)) <- names(chunk_dims_obs)
      if (dim(new_data_obs)[which(names(chunk_dims_obs) == 'time')] > (largest_ftime * (i - 1) + ftime_amount)) {
        step_new_data_exp <- abind(step_new_data_exp, 
                               Subset(new_data_obs, 'time', (largest_ftime * (i - 1) + ftime_amount + 1):
                                                            (dim(new_data_obs)[which(names(chunk_dims_obs) == 'time')])),
                               along = which(names(chunk_dims_obs) == 'time'))
        names(dim(step_new_data_exp)) <- names(chunk_dims_obs)
      }
      new_data_obs <- step_new_data_exp
      rm(step_new_data_exp)
      gc()
  #    new_dates$start <- c(head(new_dates$start, largest_ftime * (i - 1) + ftime_amount),
  #                         rep(NA, largest_ftime - ftime_amount),
  #                         new_dates$start[(largest_ftime * (i - 1) + ftime_amount + 1):
  #                                         (dim(new_data_obs)[which(names(chunk_dims_obs) == 'time')])]
  #                         )
  #    new_dates$end <- c(head(new_dates$end, largest_ftime * (i - 1) + ftime_amount),
  #                       rep(NA, largest_ftime - ftime_amount),
  #                       new_dates$end[(largest_ftime * (i - 1) + ftime_amount + 1):
  #                                       (dim(new_data_obs)[which(names(chunk_dims_obs) == 'time')])]
  ##                       tail(new_dates$end, length(new_dates$end) - (largest_ftime * (i - 1) + ftime_amount))
  #                       )
    }
  }
  # Defining exp new dims
  new_dims <- dim(data_exp$Data)
  time_dim <- which(names(new_dims) == 'time')
  before_dims <- NULL
  if (time_dim > 1) {
    before_dims <- new_dims[1:(time_dim - 1)]
  }
  after_dims <- NULL
  if (time_dim < length(new_dims)) {
    after_dims <- new_dims[(time_dim + 1):length(new_dims)]
  }
  new_dims <- c(before_dims, largest_ftime, length(data_exp$InitializationDates), after_dims)
  names(new_dims)[time_dim:(time_dim + 1)] <- c('ftime', 'sdate')
  new_dims <- c(c(dataset = 1), new_dims)
  dim(new_data_exp) <- new_dims
  perm <- 1:length(new_dims)
  perm[c(time_dim, time_dim + 1) + 1] <- perm[c(time_dim + 1, time_dim) + 1]
  new_data_exp <- aperm(new_data_exp, perm)
  # Defining obs new dims
  new_dims['member'] <- 1
  # Defining date new dims
  dim(new_dates$start) <- dim(new_data_exp)[c('ftime', 'sdate')]
  new_dates$start <- t(new_dates$start)
  names(dim(new_dates$start)) <- c('sdate', 'ftime')
  dim(new_dates$end) <- dim(new_data_exp)[c('ftime', 'sdate')]
  new_dates$end <- t(new_dates$end)
  names(dim(new_dates$end)) <- c('sdate', 'ftime')
  dim(new_data_obs) <- new_dims
  new_data_obs <- aperm(new_data_obs, perm)
  # Casting output
  data_exp_bk <- data_exp
  data_obs_bk <- data_obs
  data_exp <- list()
  data_exp$mod <- new_data_exp
  data_exp$obs <- new_data_obs
  data_exp$Dates <- new_dates
  data_exp$lon <- data_exp_bk$xyCoords$x
  data_exp$lat <- data_exp_bk$xyCoords$y
  data_exp$InitializationDates <- data_exp_bk$InitializationDates
  data_exp$Variable <- data_exp_bk$Variable
  data_exp$Members <- data_exp_bk$Members
  attributes(data_exp) <- c(attributes(data_exp), 
                        attributes(data_exp_bk)[-which(names(attributes(data_exp_bk)) %in% c('names', 'dataset'))])
  attributes(data_obs) <- c(attributes(data_obs), 
                            attributes(data_obs_bk)[-which(names(attributes(data_obs_bk)) %in% c('names', 'dataset'))])
  attr(data_exp$mod, 'dataset') <- attr(data_exp_bk, 'dataset')
  attr(data_exp$obs, 'dataset') <- attr(data_obs_bk, 'dataset')

  data_exp
}
